###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.datatypes.date import Date
from onyx.database.objdb import ObjDbDummyClient, ObjNotFound
from onyx.database.objdb_api import (UseDatabase, ExistsInDatabase, ObjDbQuery,
                                     AddObj, GetObj, UpdateObj, DelObj)

from onyx.database.test import ufo_testcls

import unittest
import os


# --- test API with a dummy client
class Test_Dummy(unittest.TestCase):
    def setUp(self):
        self.obj = ufo_testcls.ufocls(Name="test")
        self.clt = ObjDbDummyClient("dummy_db", "dummy_user")

    def test_add_get_update_delete(self):
        with UseDatabase(self.clt):
            self.assertRaises(ObjNotFound, GetObj, self.obj.Name, False)
            self.assertEqual(AddObj(self.obj), self.obj)
            self.assertEqual(GetObj(self.obj.Name), self.obj)

            self.obj.Birthday = Date.now()
            self.assertEqual(UpdateObj(self.obj), self.obj)

            DelObj(self.obj)
            self.assertRaises(ObjNotFound, GetObj, self.obj.Name)

    def test_exists(self):
        with UseDatabase(self.clt):
            self.assertFalse(ExistsInDatabase(self.obj.Name))
            AddObj(self.obj)
            self.assertTrue(ExistsInDatabase(self.obj.Name))
            DelObj(self.obj)
            self.assertFalse(ExistsInDatabase(self.obj.Name))


# --- test API with the postgres client on TestDb
class Test_Postgres(unittest.TestCase):
    def setUp(self):
        self.obj = ufo_testcls.ufocls(Name="test")
        self.clt = os.getenv("OBJDB_TEST", default="TestDb")

    def test_add_get_update_delete(self):
        with UseDatabase(self.clt):
            try:
                self.assertRaises(ObjNotFound, GetObj, self.obj.Name, False)
                self.assertEqual(AddObj(self.obj), self.obj)

                self.assertEqual(GetObj(self.obj.Name, True), self.obj)
                self.assertEqual(GetObj(self.obj.Name, True).Birthday,
                                 self.obj.Birthday)
                self.assertEqual(GetObj(self.obj.Name, True).OtherDates,
                                 self.obj.OtherDates)
                self.assertEqual(GetObj(self.obj.Name, True).SimpleCurve,
                                 self.obj.SimpleCurve)

                self.obj.Birthday = Date.now()
                self.assertEqual(UpdateObj(self.obj), self.obj)

                self.assertEqual(GetObj(self.obj.Name, True), self.obj)

                DelObj(self.obj)
                self.assertRaises(ObjNotFound, GetObj, self.obj.Name)

            finally:
                # --- try cleaning up irrespective of failures
                ObjDbQuery("DELETE FROM Objects "
                           "WHERE Name=%s", (self.obj.Name,))
                ObjDbQuery("DELETE FROM ClassDefinitions "
                           "WHERE ObjType=%s", (self.obj.ObjType,))

    def test_exists(self):
        with UseDatabase(self.clt):
            try:
                self.assertFalse(ExistsInDatabase(self.obj.Name))
                AddObj(self.obj)
                self.assertTrue(ExistsInDatabase(self.obj.Name))
                DelObj(self.obj)
                self.assertFalse(ExistsInDatabase(self.obj.Name))

            finally:
                # --- try cleaning up irrespective of failures
                ObjDbQuery("DELETE FROM Objects "
                           "WHERE Name=%s", (self.obj.Name,))
                ObjDbQuery("DELETE FROM ClassDefinitions "
                           "WHERE ObjType=%s", (self.obj.ObjType,))

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.database.ufo_base import UfoBase
from onyx.depgraph.graph import GraphNodeError, GraphNodeCalc
from onyx.depgraph.graph_api import GetNode

import onyx.depgraph as onyx_dg

__all__ = ["SetDiddle", "EvalBlock", "DiddleScope"]


##-----------------------------------------------------------------------------
def SetDiddle(obj_name, VT, value, eval_block=None):
    """
    Description:
        Diddle the value of a ValueType. This is generally used withing an
        EvalBlock.
        NB: only Instreams and VTs defined as Property can be diddled.
    Inputs:
        obj_name   - target object in database (or memory)
        VT         - target value type (an Instream or a method decorated by
                     @GraphNodeVt)
        value      - the value to diddle the VT to
        eval_block - EvalBlock context manager
    Returns:
        None.
    """
    # --- make sure that obj is the object's Name
    if isinstance(obj_name, UfoBase):
        obj_name = obj_name.Name

    # --- get the target node from the graph
    name = (obj_name, VT)
    node = GetNode(name)

    # --- raise exception if VT cannot be diddled
    if isinstance(node, GraphNodeCalc):
        raise GraphNodeError("%s, %s: Only Instreams "
                             "and Properties can be diddled".format(name))

    # --- look for a valid (i.e. not used) node name for the old node
    idx = 0
    name_old = (obj_name, VT, idx)
    while name_old in onyx_dg.graph:
        idx += 1
        name_old = (obj_name, VT, idx)

    # --- clone original node and keep it on graph as name_old
    onyx_dg.graph[name_old] = node.clone()

    # --- save the name of the original node
    node.old_node = name_old

    # --- remove this node from the set of parents of all its children
    for child in node.children:
        onyx_dg.graph[child].parents.remove(name)

    # --- remove all its children
    node.children = set()

    # --- invalidate all its parents
    for parent in node.parents:
        onyx_dg.graph[parent].invalidate()

    # --- set the node value and set its state to valid
    node.value = value
    node.valid = True

    # --- if present, add name of diddled node to the EvalBlock context manager
    if eval_block is not None:
        eval_block.append(name)


###############################################################################
class EvalBlock(object):
    """
    Description:
        Context manager used to manage lifetime of one or more one-off diddles.
    Usage:
        
    """
    ##-------------------------------------------------------------------------
    def __init__(self):
        self.__diddles = []

    def append(self, value):
        self.__diddles.append(value)

    ##-------------------------------------------------------------------------
    def __enter__(self):
        # --- return a reference to itself (to be used by SetDiddle)
        return self

    ##-------------------------------------------------------------------------
    def __exit__(self, type, value, traceback):
        for obj_name, VT in reversed(self.__diddles):
            name = (obj_name, VT)
            name_old = onyx_dg.graph[name].old_node

            # --- invalidate all parents
            for parent in (onyx_dg.graph[name].parents |
                           onyx_dg.graph[name_old].parents):
                onyx_dg.graph[parent].invalidate()

            # --- reconstruct connection with children
            for child in onyx_dg.graph[name_old].children:
                onyx_dg.graph[child].parents.add(name)

            # --- recover original node from the graph
            onyx_dg.graph[name] = onyx_dg.graph[name_old]

            # --- remove original node from the graph
            del onyx_dg.graph[name_old]


###############################################################################
class DiddleScope(object):
    """
    Description:
        Context manager used to manage lifetime of diddles that will be used
        several times.
    Usage:
        
    """
    ##-------------------------------------------------------------------------
    def __init__(self):
        self.__active = False
        self.__diddles = []

    ##-------------------------------------------------------------------------
    def set_diddle(self, obj_name, VT, value):
        if isinstance(obj_name, UfoBase):
            obj_name = obj_name.Name
        if self.__active:
            SetDiddle(obj_name, VT, value)
        self.__diddles.append((obj_name, VT, value))

    ##-------------------------------------------------------------------------
    def __enter__(self):
        for obj, VT, value in self.__diddles:
            SetDiddle(obj, VT, value)
        self.__active = True
        return self

    ##-------------------------------------------------------------------------
    def __exit__(self, type, value, traceback):
        for obj_name, VT, _ in reversed(self.__diddles):
            name = (obj_name, VT)
            name_old = onyx_dg.graph[name].old_node

            # --- invalidate all parents
            for parent in (onyx_dg.graph[name].parents |
                           onyx_dg.graph[name_old].parents):
                onyx_dg.graph[parent].invalidate()

            # --- reconstruct connection with children
            for child in onyx_dg.graph[name_old].children:
                onyx_dg.graph[child].parents.add(name)

            # --- recover original node from the graph
            onyx_dg.graph[name] = onyx_dg.graph[name_old]

            # --- remove original node from the graph
            del onyx_dg.graph[name_old]

        self.__active = False
        return False



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.datatypes.date import Date

import datetime
import unittest
import pickle


# --- Unit tests
class RegTest(unittest.TestCase):
    def setUp(self):
        # --- perform set-up actions, if any
        pass

    def tearDown(self):
        # perform clean-up actions, if any
        pass

    def test_constructors(self):
        self.assertEqual(Date.today(), datetime.date.today())
        self.assertLessEqual(Date.now(), datetime.datetime.now())
        ref = "01-Jan-2008"
        self.assertEqual(Date(2008, 1, 1).__str__(), ref)
        self.assertEqual(Date.parse("01Jan08").__str__(), ref)
        self.assertEqual(Date.parse("01 Jan 08").__str__(), ref)
        self.assertEqual(Date.parse("01-Jan-08").__str__(), ref)
        self.assertEqual(Date.parse("F08").__str__(), ref)
        # --- test conversion to float and back
        ref = Date(2008, 1, 1)
        self.assertEqual(Date.parse(ref.ordinal), ref)
        ref = Date(2008, 1, 1, 10, 33, 48, 0)
        self.assertEqual(Date.parse(ref.ordinal), ref)

    def test_date_algebra(self):
        d0 = Date(2008, 1, 1)
        d1 = Date(2008, 12, 31)
        dt = d1 - d0
        self.assertEqual(dt.days, 366 - 1)
        d0 = Date(2009, 1, 1)
        d1 = Date(2009, 12, 31)
        dt = d1 - d0
        self.assertEqual(dt.days, 365 - 1)
        # --- date comparisons
        self.assertTrue(Date(2008, 1, 1) < Date(2009, 1, 1))
        self.assertFalse(Date(2008, 1, 1) == Date(2009, 1, 1))
        self.assertTrue(Date(2008, 1, 1) != Date(2009, 1, 1))
        self.assertFalse(Date(2008, 1, 1) > Date(2009, 1, 1))
        # --- now between Date and datetime objects
        self.assertTrue(Date(2008, 1, 1) == datetime.date(2008, 1, 1))
        self.assertTrue(Date(2009, 1, 1) != datetime.date(2008, 1, 1))
        self.assertTrue(Date(2009, 1, 1) > datetime.date(2008, 1, 1))
        self.assertTrue(Date(2008, 1, 1) == datetime.datetime(2008, 1, 1))
        self.assertTrue(Date(2009, 1, 1) != datetime.datetime(2008, 1, 1))
        self.assertTrue(Date(2009, 1, 1) > datetime.datetime(2008, 1, 1))

    def test_pickling(self):
        d0 = Date(1977, 6, 8)
        d1 = pickle.loads(pickle.dumps(d0, 2))
        # --- check that data representations are identical
        self.assertEqual(d0, d1)
        # --- check that string representations are identical
        self.assertEqual(d0.__str__(), d1.__str__())

    def test_is_weekday(self):
        from calendar import THURSDAY
        self.assertFalse(Date(2008, 1, 1).is_weekday(THURSDAY))
        self.assertTrue(Date(2009, 1, 1).is_weekday(THURSDAY))

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.datatypes.date import Date, DateError, L2M
from onyx.datatypes.rdate import RDate

__all__ = [
    "LYY2Date",
    "Date2LYY",
    "Date2QQYY",
    "DateRange",
    "CalcTerm",
    "CountBizDays",
]

# --- auxiliary local variables (not exported)
M2L = dict(zip(L2M.values(), L2M.keys()))
QUARTER_BY_MTH = [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4]


##-----------------------------------------------------------------------------
def LYY2Date(lyy):
    """
    Description:
        Convert a Nymex LYY string (used for futures contracts as in F09) to
        a Date. The LYY string is assumed to correspond to the first day in the
        corresponding month.
    Inputs:
        lyy - a string defining a valid Nymex future contract.
    Returns:
        A Date.
    """
    try:
        m = L2M[lyy[0]]
        y = int(lyy[1:])
        y = 2000 + y if y < 70 else 1900 + y
        return Date(y, m, 1)
    except KeyError:
        raise DateError("Unrecognized LYY string: {0:s}".format(lyy))


##-----------------------------------------------------------------------------
def Date2LYY(date):
    """
    Description:
        Convert a Date to a Nymex LYY string (used for futures contracts
        as in F09). Conversion is done ignoring the actual day in the month.
    Inputs:
        date - a Date for which the LYY Nymex future contract is returned.
    Returns:
        A string with the LYY code.
    """
    if date < Date.low_date() or date > Date.high_date():
        raise DateError("Date outside valid range: {0:s}".format(date))

    m = date.month
    y = date.year
    yy = y - 2000 if y > 1999 else y - 1900

    return "{0:1s}{1:02d}".format(M2L[m], yy)


##-----------------------------------------------------------------------------
def Date2QQYY(date):
    """
    Description:
        Convert a Date to a string for the corresponding quarter (as in Q311).
        Conversion is done ignoring the actual day in the quarter.
    Inputs:
        date - a Date for which the corresponding quarter is returned.
    Returns:
        A string with the QQYY code.
    """
    if date < Date.low_date() or date > Date.high_date():
        raise DateError("Date outside valid range: {0:s}".format(date))

    m = date.month - 1
    y = date.year
    yy = y - 2000 if y > 1999 else y - 1900

    return "Q{0:1d}{1:02d}".format(QUARTER_BY_MTH[m], yy)


##-----------------------------------------------------------------------------
def DateRange(start, end, rule, calendar=None):
    """
    Description:
        Iterator that yields a series of dates between start date and
        end date according to increment rule (start date is included).
    Inputs:
        start    - the first date in the range
        end      - the last  date in the range
        rule     - a date rule (see RDate)
        calendar - a holiday calendar used to identify business days
    Yields:
        A range of Dates.
    """
    d = start
    rd = RDate(rule, calendar)
    while d <= end:
        yield d
        d = d + rd


##-----------------------------------------------------------------------------
def CalcTerm(start, end, base=365):
    """
    Description:
        Return the number of years between start and end date assuming that
        each year has a number of days as specified by base. Term is negative
        if start > end.
    Inputs:
        start - the initial date of the term
        end   - the final   date of the term
        base  - number of days in a calendar year
    Returns:
        A float.
    """
    return (end - start).days / float(base)


##-----------------------------------------------------------------------------
def CountBizDays(start, end, calendar=None):
    """
    Description:
        Return the number of business days between start and end date according
        to the holiday calendar.
        NB: if start date == end date this function will return 1.
    Inputs:
        start    - the initial date of the range
        end      - the last  date in the range
        calendar - a holiday calendar used to identify business days
    Returns:
        An integer.
    """
    # --- first push start date forward to the nearest business day
    start = start + RDate("+0b", calendar)
    # --- then sum all business days in the range
    return sum((1 for d in DateRange(start, end, "+1b", calendar)))



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.datatypes.structure import Structure

import unittest
import pickle


# --- Unit tests
class RegTest(unittest.TestCase):
    def setUp(self):
        # --- perform set-up actions, if any
        pass

    def tearDown(self):
        # --- perform clean-up actions, if any
        pass

    def test_constructors(self):
        # --- test different ways of creating the same structure
        # --- first: this doesn't preserve the order!!!
        struct1 = Structure({"a": 1, "b": 2, "c": 3})
        # --- second: this doesn't preserve the order!!!
        struct2 = Structure(a=1, b=2, c=3)
        # --- third: this preserves the order
        struct3 = Structure([("a", 1), ("b", 2), ("c", 3)])
        # --- fourth: this preserves the order
        struct4 = Structure()
        struct4["a"] = 1
        struct4["b"] = 2
        struct4["c"] = 3

        self.assertEqual(struct1, struct2)
        self.assertEqual(struct2, struct3)
        self.assertEqual(struct3, struct4)

    def test_algerba(self):
        struct1 = Structure([('a',  1), ('b',  2), ('c',  3)])
        struct2 = Structure([('b', 10), ('c', 10), ('d', 10)])

        add_struct = Structure([('a', 1), ('b', 12), ('c', 13), ('d', 10)])
        sub_struct = Structure([('a', -1), ('b', 8), ('c', 7), ('d', 10)])

        self.assertEqual(struct2 + struct1, add_struct)
        self.assertEqual(struct2 - struct1, sub_struct)

        iadd_struct = Structure([('a', 1), ('b', 12), ('c', 13), ('d', 10)])

        struct2 += struct1
        self.assertEqual(struct2, iadd_struct)

        rmul_struct = Structure([('a', -2), ('b', -4), ('c', -6)])

        self.assertEqual(-2*struct1, rmul_struct)

    def test_pickling(self):
        struct0 = Structure({"a": 1, "b": 2, "c": 3})
        struct1 = pickle.loads(pickle.dumps(struct0, 2))
        # --- check that data representations are identical
        self.assertEqual(struct0, struct1)
        # --- check that string representations are identical
        self.assertEqual(struct0.__str__(), struct1.__str__())

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.datatypes.curve import Curve
from onyx.datatypes.hlocv import HlocvCurve
from onyx.database.tsdb import TsDbClient, TsDbError

import onyx.database as onyx_db

from psycopg2.extensions import (ISOLATION_LEVEL_AUTOCOMMIT,
                                 ISOLATION_LEVEL_READ_COMMITTED)

import getpass

__all__ = [
    "TsDbUseDatabase",
    "TsDbTransaction",
    "TsDbQuery",
    "TsDbGetRowBy",
    "TsDbUpsertCurve",
    "TsDbGetCurve",
]

USER = getpass.getuser()


###############################################################################
class TsDbUseDatabase(object):
    """
    Context manager used to setup the tsdb client used by the api.
    """
    def __init__(self, dbname):
        self.dbname = dbname

    def __enter__(self):
        self.current = onyx_db.ts_clt
        onyx_db.ts_clt = TsDbClient(self.dbname, USER)

    def __exit__(self, exc_type, exc_value, traceback):
        onyx_db.ts_clt.close()
        onyx_db.ts_clt = self.current
        # --- returns False so that all execptions raised will be propagated
        return False


###############################################################################
class TsDbTransaction(object):
    """
    Transaction context manager.
    """
    def __init__(self, name):
        self.name = name

    def __enter__(self):
        onyx_db.ts_clt.conn.set_isolation_level(ISOLATION_LEVEL_READ_COMMITTED)
        onyx_db.ts_clt.conn.tpc_begin(self.name)

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            onyx_db.ts_clt.conn.tpc_commit()
        else:
            onyx_db.ts_clt.conn.tpc_rollback()
        # --- restore isolation level
        onyx_db.ts_clt.conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        # --- returns False so that all execptions raised will be propagated
        return False


##-----------------------------------------------------------------------------
def TsDbQuery(query, parms=None, attr=""):
    """
    Description:
        Perform a db-query.
    Inputs:
        query - the SQL query as a valid string.
        parms - parameters, if any, that are passed to the query
        attr  - cursor attribute, such as fetchone or fetchall.
    Returns:
        The result of the query.
    """
    try:
        curs = onyx_db.ts_clt.conn.cursor()
    except AttributeError:
        raise TsDbError("{0:s} does not support "
                        "queries".format(onyx_db.ts_clt.dbname))

    try:
        if parms is None:
            curs.execute(query)
        else:
            curs.execute(query, parms)
        return getattr(curs, attr, lambda: None)()
    finally:
        curs.close()


##-----------------------------------------------------------------------------
def TsDbGetRowBy(table, name, date=None, strict=False):
    """
    Description:
        Fetch a a single row from database.
    Inputs:
        table  - the table in TsDb
        name   - the name of the time series as stored in database
        date   - the required date (if None fetch the most recend row)
        strict - if True, raise exception if row is not found. Otherwise
                 return most recent matching row.
    Returns:
        The database row.
    """
    return onyx_db.ts_clt.get_row_by(table, name, date, strict)


##-----------------------------------------------------------------------------
def TsDbUpsertCurve(name, crv):
    """
    Description:
        Add/update curve to the database.
    Inputs:
        name - the name of the time series as stored in database
        crv  - the curve to upsert
    Returns:
        None.
    """
    if isinstance(crv, HlocvCurve):
        onyx_db.ts_clt.upsert_curve(name, crv, "HLOCV")
    elif isinstance(crv, Curve):
        onyx_db.ts_clt.upsert_curve(name, crv, "CRV")
    else:
        raise ValueError("Unrecognized curve type: {0:s}".format(type(crv)))


##-----------------------------------------------------------------------------
def TsDbGetCurve(name, start=None, end=None, crv_type="CRV", field=None):
    """
    Description:
        Fetch a plain time series from to the database.
    Inputs:
        name     - the name of the time series as stored in database
        start    - start date, inclusive
        end      - end date, inclusive
        crv_type - choose between CRV and HLOCV
        field    - for HLOCV curves, select the specific field to return
                   among: High, Low, Open, Close, Volume
                   If None, and HLOCV curve is returned instead.
    Returns:
        The time series as a curve.
    """
    if crv_type == "CRV":
        return onyx_db.ts_clt.get_curve(name, start, end)
    elif crv_type == "HLOCV":
        return onyx_db.ts_clt.get_hlocv(name, start, end, field)
    else:
        raise ValueError("Unrecognized curve type: {0:s}".format(crv_type))



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.datatypes.gcurve import GKnot, GCurve

import numpy as np
import bisect

__all__ = ["Knot", "Curve"]


##-----------------------------------------------------------------------------
def is_numlike(obj):
    """
    Helper function used to determine if an object looks like a number.
    """
    try:
        obj + 1
    except:
        return False
    else:
        return True


###############################################################################
class Knot(GKnot):
    """
    Class representing a curve knot as a pair (date, value) where value is
    stored as a float datatype.
    """
    __slots__ = ("date", "value")

    ##-------------------------------------------------------------------------
    def __init__(self, d, v):
        self.date = d
        self.value = float(v)

    ##-------------------------------------------------------------------------
    def __str__(self):
        return "{0!s}, {1:17.8f}".format(self.date, self.value)


###############################################################################
class Curve(GCurve):
    """
    Curve class (based on Knot class, which only stores numeric values)
    """
    __slots__ = ("data", "knot_cls")

    ##-------------------------------------------------------------------------
    def __init__(self, dates=None, values=None, knot_cls=None):
        """
        Description:
            Create a curve from dates and numeric values. Raise a CurveError
            exception if dates contains any duplicates.
        Inputs:
            dates    - list/array of Dates
            values   - list/array of values (values must be castable to float)
            knot_cls - optional, the Knot constructor
        """
        knot_cls = knot_cls or Knot
        super().__init__(dates, values, knot_cls)

    ##-------------------------------------------------------------------------
    @property
    def values(self):
        return np.fromiter((knot.value for knot in self.data),
                           dtype=float, count=len(self.data))

    ##-------------------------------------------------------------------------
    def __getitem__(self, d):
        """
        Reimplement __getitem__ using linear interpolation.
        """
        dts = self.dates
        idx = bisect.bisect_left(dts, d)

        if d >= dts[0] and d <= dts[-1] and dts[idx] == d:
            return self.data[idx].value
        else:
            # --- use linear interpolation
            if idx == 0:
                return self.data[0].value
            elif idx == len(dts):
                return self.data[-1].value
            else:
                frc = ((d.ordinal - dts[idx-1].ordinal) /
                       (dts[idx].ordinal - dts[idx-1].ordinal))
                cfrc = 1.0 - frc
                return (cfrc*self.data[idx-1].value + frc*self.data[idx].value)

    ##-------------------------------------------------------------------------
    ##  A few methods for simple curve algebra. It's always assumed that the
    ##  two curves are correctly aligned

    def __add__(self, other):
        if isinstance(other, self.__class__):
            return self.__class__(self.dates, self.values + other.values)
        else:
            return self.__class__(self.dates, self.values + other)

    def __sub__(self, other):
        if isinstance(other, self.__class__):
            return self.__class__(self.dates, self.values - other.values)
        else:
            return self.__class__(self.dates, self.values - other)

    def __mul__(self, other):
        if isinstance(other, self.__class__):
            return self.__class__(self.dates, self.values * other.values)
        elif is_numlike(other):
            return self.__class__(self.dates, self.values * other)
        else:
            TypeError("cannot multiply a {0:s} "
                      "by a {1:s}".format(self.__class__, other.__class__))

    def __truediv__(self, other):
        if isinstance(other, self.__class__):
            return self.__class__(self.dates, self.values / other.values)
        elif is_numlike(other):
            return self.__class__(self.dates, self.values / other)
        else:
            TypeError("cannot multiply a {0:s} "
                      "by a {1:s}".format(self.__class__, other.__class__))

    def __radd__(self, scalar):
        return self.__class__(self.dates, scalar + self.values)

    def __rsub__(self, scalar):
        return self.__class__(self.dates, scalar - self.values)

    def __rmul__(self, scalar):
        return self.__class__(self.dates, scalar * self.values)

    def __rtruediv__(self, scalar):
        return self.__class__(self.dates, scalar / self.values)

    ##-------------------------------------------------------------------------
    def crop_values(self, start=None, end=None):
        """
        Description:
            Return an array of values that fall within the specified range.
            If start date or end date are not set, use the current extremes of
            the curve.
        Inputs:
            start - the start of the range (included)
            end   - the   end of the range (included)
        Returns:
            An array of values within the range.
        """
        # --- store dates in local variable for faster access
        dts = self.dates

        # --- set the two extremes of the range
        start = start or dts[0]
        end = end or dts[-1]

        i = bisect.bisect_left(dts, start)
        j = bisect.bisect_right(dts, end, i)

        return self.values[i:j]



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import numpy as np
import collections
import bisect

__all__ = ["GKnot", "GCurve", "KnotError", "CurveError"]


###############################################################################
class KnotError(Exception):
    """
    Base class for all Knot exceptions.
    """
    pass


###############################################################################
class CurveError(Exception):
    """
    Base class for all Curve exceptions.
    """
    pass


###############################################################################
class GKnot(object):
    """
    Class representing a generic curve knot as a pair (date, value).
    """
    __slots__ = ("date", "value")

    ##-------------------------------------------------------------------------
    def __init__(self, d, v):
        self.date = d
        self.value = v

    ##-------------------------------------------------------------------------
    ##  support serialization

    def __getstate__(self):
        return (self.date, self.value)

    def __setstate__(self, state):
        self.date, self.value = state

    ##-------------------------------------------------------------------------
    def __repr__(self):
        cls = self.__class__
        return "{.__name__:s}({.date!r},{.value!r})".format(cls, self, self)

    ##-------------------------------------------------------------------------
    def __str__(self):
        return "{0!s}, unprintable value...".format(self.date)

    ##-------------------------------------------------------------------------
    ##  methods for knot comparison (date based)

    def __lt__(self, other):
        return self.date < other.date

    def __le__(self, other):
        return self.date <= other.date

    def __gt__(self, other):
        return self.date > other.date

    def __ge__(self, other):
        return self.date >= other.date

    ##-------------------------------------------------------------------------
    ##  quality/inquality are based on both date and value

    def __eq__(self, other):
        return self.date == other.date and self.value == other.value

    def __ne__(self, other):
        return self.date != other.date or self.value != other.value


###############################################################################
class GCurve(object):
    """
    Base Curve class (based on GKnot class).
    """
    __slots__ = ("data", "knot_cls")

    ##-------------------------------------------------------------------------
    def __init__(self, dates=None, values=None, knot_cls=None):
        """
        Description:
            Create a curve from dates and generic values. Raise a CurveError
            exception if dates contains any duplicates.
        Inputs:
            dates    - list/array of Dates
            values   - list/array of values
            knot_cls - optional Knot constructor
        """
        if dates is None: dates = []    # analysis:ignore
        if values is None: values = []  # analysis:ignore

        if len(dates) != len(values):
            raise CurveError("Input dates and values are of different length")

        self.knot_cls = knot_cls or GKnot

        # --- check if creating an empty curve
        if not len(dates):
            self.data = []
        else:
            # --- check whether there are any duplicates in dates
            if len(dates) != len(set(dates)):
                # --- raise exception with list of duplicates
                duplicates = {x: y for x, y
                              in collections.Counter(dates).items() if y > 1}
                raise CurveError("Creating curve with duplicate "
                                 "dates: {0:s}".format(duplicates.__repr__()))

            # --- the curve is stored internally as a list of knots
            self.data = [self.knot_cls(d, v) for (d, v) in zip(dates, values)]
            # --- sort the curve by dates (natural ordering of knots)
            self.data.sort()

    ##-------------------------------------------------------------------------
    ##  methods for curve comparison

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        for i in range(len(self)):
            if self.data[i] != other.data[i]:
                return False
        return True

    def __ne__(self, other):
        if len(self) != len(other):
            return True
        for i in range(len(self)):
            if self.data[i] != other.data[i]:
                return True
        return False

    ##-------------------------------------------------------------------------
    def sort(self, *args, **kwds):
        self.data.sort(*args, **kwds)

    ##-------------------------------------------------------------------------
    ##  support serialization

    def __getstate__(self):
        return self.knot_cls, self.data

    def __setstate__(self, state):
        self.knot_cls, self.data = state

    ##-------------------------------------------------------------------------
    ##  string representation

    def __repr__(self):
        cls = self.__class__.__name__
        dts = self.dates
        vls = list(self.values)
        return "{0:s}(dates={1!r}, values={2!r})".format(cls, dts, vls)

    def __str__(self):
        return "\n".join([knot.__str__() for knot in self.data])

    ##-------------------------------------------------------------------------
    def has_knot(self, d):
        """
        Description:
            Check whether a knot for a given date is present.
        Inputs:
            d - the knot date
        Returns:
            True if knot is present, False otherwise.
        """
        dts = self.dates
        if not len(dts):
            return False
        if d < dts[0] or d > dts[-1]:
            return False
        idx = bisect.bisect_left(dts, d)
        if dts[idx] == d:
            return True
        else:
            return False

    ##-------------------------------------------------------------------------
    def get_knot(self, d):
        """
        Description:
            Return the knot for a given date if present. Raise IndexError
            exception otherwise.
        Inputs:
            d - the knot date
        Returns:
            The curve's knot.
        """
        dts = self.dates
        idx = bisect.bisect_left(dts, d)
        try:
            if dts[idx] == d:
                return self.data[idx]
            else:
                raise IndexError
        except:
            raise IndexError

    ##-------------------------------------------------------------------------
    def add_knot(self, knot):
        """
        Description:
            Add a knot in-place to the curve preserving order.
        Inputs:
            Knot - the knot to be added to the curve
        Returns:
            None.
        """
        bisect.insort(self.data, knot)

    ##-------------------------------------------------------------------------
    def del_knot(self, d):
        """
        Description:
            Delete in-place the knot corresponding to a given date if the knot
            is present. Raise IndexError exception otherwise.
        Inputs:
            d - the knot date
        Returns:
            None.
        """
        dts = self.dates
        idx = bisect.bisect_left(dts, d)
        try:
            if dts[idx] == d:
                self.data.pop(idx)
            else:
                raise IndexError
        except:
            raise IndexError

    ##-------------------------------------------------------------------------
    ##  GCurve is an iterable object: implement all relevant methods

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def __getitem__(self, d):
        dts = self.dates
        idx = bisect.bisect_left(dts, d)

        if d >= dts[0] and d <= dts[-1] and dts[idx] == d:
            return self.data[idx].value
        else:
            # --- use step interpolation (previous knot)
            if idx:
                return self.data[idx-1].value
            else:
                return self.data[0].value

    def __setitem__(self, d, v):
        try:
            idx = self.dates.index(d)
            self.data[idx].value = v
        except ValueError:
            bisect.insort(self.data, self.knot_cls(d, v))

    ##-------------------------------------------------------------------------
    def append(self, knot):
        """
        Append new knot (assuming it's already in the correct position).
        """
        self.data.append(knot)

    ##-------------------------------------------------------------------------
    @property
    def ordinals(self):
        """
        Return an array with curve dates in numeric format.
        """
        return np.fromiter((knot.date.ordinal for knot in self.data),
                           dtype=float, count=len(self.data))

    ##-------------------------------------------------------------------------
    @property
    def dates(self):
        """
        Return a list with curve dates in Date format.
        """
        return [knot.date for knot in self.data]

    ##-------------------------------------------------------------------------
    @property
    def values(self):
        """
        Return a list with curve's knot values.
        """
        return [knot.value for knot in self.data]

    ##-------------------------------------------------------------------------
    @property
    def front(self):
        """
        Description:
            Return the first knot of the curve.
        Returns:
            A knot.
        """
        return self.data[0]

    @property
    def back(self):
        """
        Description:
            Return the last knot of the curve.
        Returns:
            A knot.
        """
        return self.data[-1]

    ##-------------------------------------------------------------------------
    def crop(self, start=None, end=None):
        """
        Description:
            Return a sub-curve with knots that fall within the specified range.
            If start date or end date are not set, use the current extremes of
            the curve.
        Inputs:
            start - the start of the range (included)
            end   - the   end of the range (included)
        Return:
            A sub-curve of the original curve.
        """
        # --- store dates in local variable for faster access
        dts = self.dates

        # --- set the two extremes of the range
        start = start or dts[0]
        end = end or dts[-1]

        i = bisect.bisect_left(dts, start)
        j = bisect.bisect_right(dts, end, i)

        return self.__class__(dts[i:j], self.values[i:j])



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import collections
import copy

__all__ = ["Structure", "StructureError"]


###############################################################################
class StructureError(Exception):
    """
    Base class for all Structure exceptions.
    """
    pass


###############################################################################
class Structure(collections.OrderedDict):
    """
     Inherits from ordered dictionary and implements addition and subtraction
     functionality.
    """
    ##-------------------------------------------------------------------------
    ##  addition methods

    def __iadd__(self, other):
        for key, value in other.items():
            self[key] = self.get(key, 0.0) + value
        return self

    def __add__(self, other):
        struct = copy.deepcopy(self)
        for key, value in other.items():
            struct[key] = struct.get(key, 0.0) + value
        return struct

    ##-------------------------------------------------------------------------
    ##  subtraction methods

    def __isub__(self, other):
        for key, value in other.items():
            self[key] = self.get(key, 0.0) - value
        return self

    def __sub__(self, other):
        struct = copy.deepcopy(self)
        for key, value in other.items():
            struct[key] = struct.get(key, 0.0) - value
        return struct

    ##-------------------------------------------------------------------------
    ##  addition and multiplication by scalars

    def __radd__(self, scalar):
        struct = copy.deepcopy(self)
        for key in struct:
            struct[key] += scalar
        return struct

    def __rmul__(self, scalar):
        struct = copy.deepcopy(self)
        for key in struct:
            struct[key] *= scalar
        return struct

    ##-------------------------------------------------------------------------
    ##  equality, order is not tested
    def __eq__(self, other):
        return dict.__eq__(self, other)

    ##-------------------------------------------------------------------------
    def drop_zeros(self):
        """
        Drop items with zero value.
        """
        for key in self:
            if self[key] == 0.0:
                del self[key]



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.depgraph.graph_api import CreateInMemory, GetVal
from onyx.depgraph.diddles import EvalBlock, DiddleScope, SetDiddle

from onyx.database.objdb_api import UseDatabase
from onyx.depgraph.graph_api import UseGraph
from onyx.depgraph.test import ufo_testcls

import unittest
import os


class RegTest(unittest.TestCase):
    def setUp(self):
        self.use_db = UseDatabase(os.getenv("OBJDB_TEST", default="TestDb"))
        self.use_graph = UseGraph()
        self.use_db.__enter__()
        self.use_graph.__enter__()

        self.obj = CreateInMemory(ufo_testcls.TestCls3(Name="obj"))

    def tearDown(self):
        del self.obj
        self.use_db.__exit__()
        self.use_graph.__exit__()

    def node_assertEqual(self, obj, vt, value):
        self.assertEqual(GetVal(self.obj.Name, vt), value)
        self.assertEqual(GetVal(self.obj, vt), value)

    def test_evalblock(self):
        self.node_assertEqual(self.obj, "CalcProperty", 1333.0)

        # --- create an eval block
        with EvalBlock() as ev_block:
            self.node_assertEqual(self.obj, "Number", 666.0)
            self.node_assertEqual(self.obj, "CalcProperty", 1333.0)

            # --- now diddle Number to 0
            SetDiddle("obj", "Number", 0, ev_block)
            self.node_assertEqual(self.obj, "CalcProperty", 1.0)

            # --- test nested eval blocks
            with EvalBlock() as new_ev_block:
                # --- diddle Number to 333 passing new diddle scope
                SetDiddle("obj", "Number", 333, new_ev_block)
                self.node_assertEqual(self.obj, "CalcProperty", 667.0)

            # --- outside nested eval block
            self.node_assertEqual(self.obj, "CalcProperty", 1.0)

            # --- now diddle the calc VT itself
            SetDiddle("obj", "CalcProperty", 1976, ev_block)
            self.node_assertEqual(self.obj, "CalcProperty", 1976)

            with EvalBlock() as new_ev_block:
                SetDiddle("obj", "CalcProperty", 333, new_ev_block)
                self.node_assertEqual(self.obj, "CalcProperty", 333)

            # --- outside nested diddle eval block
            self.node_assertEqual(self.obj, "CalcProperty", 1976)

        # --- outside all eval blocks
        self.node_assertEqual(self.obj, "Number", 666.0)
        self.node_assertEqual(self.obj, "CalcProperty", 1333.0)

    def test_diddlescope_basicusage(self):
        self.node_assertEqual(self.obj, "Number", 666.0)
        self.node_assertEqual(self.obj, "CalcProperty", 1333.0)

        # --- create a diddle scope and set a first diddle
        scope = DiddleScope()
        scope.set_diddle(self.obj, "Number", 0.0)

        # --- the diddle is set, but not yet used
        self.node_assertEqual(self.obj, "Number", 666.0)
        self.node_assertEqual(self.obj, "CalcProperty", 1333.0)

        # --- use the diddle scope to enforce existing diddles and to add more
        with scope:
            self.node_assertEqual(self.obj, "Number", 0.0)
            self.node_assertEqual(self.obj, "CalcProperty", 1.0)

            scope.set_diddle(self.obj, "CalcProperty", "ABC")

            self.node_assertEqual(self.obj, "Number", 0.0)
            self.node_assertEqual(self.obj, "CalcProperty", "ABC")

        # --- outside diddle scope
        self.node_assertEqual(self.obj, "Number", 666.0)
        self.node_assertEqual(self.obj, "CalcProperty", 1333.0)

        # --- a diddle scope can be reused and all diddles are re-enforced
        # --- use the diddle scope to enforce existing diddles and to add more
        with scope:
            self.node_assertEqual(self.obj, "Number", 0.0)
            self.node_assertEqual(self.obj, "CalcProperty", "ABC")

        # --- outside diddle scope
        self.node_assertEqual(self.obj, "Number", 666.0)
        self.node_assertEqual(self.obj, "CalcProperty", 1333.0)

if __name__ == "__main__":
    unittest.main()



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

import operator

__all__ = ["Table", "TableError"]


###############################################################################
class TableError(Exception):
    """
    Base class for all Table exceptions.
    """
    pass


##-----------------------------------------------------------------------------
def row_factory(tags):
    """
    Table's row factory: generates a class with a predefined set of attributes.
    """
    # --- make sure that all tags are strings
    for tag in tags:
        if not isinstance(tag, str):
            raise ValueError("Table tags must be strings")

    # --- convert tags to a tuple
    tags = tuple(tags)

    # --- the appropriate row class is created on the fly
    class table_raw(object):
        __slots__ = tags

        def __init__(self, values):
            for i, value in enumerate(values):
                self.__setattr__(tags[i], value)

        def __getitem__(self, attribute):
            return self.__getattribute__(attribute)

        def __setitem__(self, attribute, value):
            self.__setattr__(attribute, value)

        def __str__(self):
            return ", ".join([str(v) for v in self.values])

        @property
        def fields(self):
            return tags

        @property
        def values(self):
            return [self.__getattribute__(a) for a in tags]

        @property
        def str_values(self):
            return [str(self.__getattribute__(a)) for a in tags]

        def __getstate__(self):
            return self.values

        def __setstate__(self, state):
            for i, attr in enumerate(tags):
                self.__setattr__(attr, state[i])

        def __eq__(self, other):
            return tags == other.__slots__ and self.values == other.values

    # --- return the newly created class
    return table_raw


###############################################################################
class Table(object):
    """
    """
    __slots__ = ("tags", "data", "row_fmt", "row_cls")

    ##-------------------------------------------------------------------------
    def __init__(self, table, col_width=20):
        """
        Description:
            Table Class: Creates a Table from an omogeneous list of tuples or
            lists.
        Input:
            table_array - an list of lists in the following format:
                          [ [ tag1, tag2, ..., tagN ],
                            [ val1, val2, ..., valN ],
                            [ ...                   ] ]
            col_width   - default column width used for printing
        """
        self.tags = tuple((str(s).replace(" ", "_") for s in table[0]))
        self.row_fmt = "".join(["{{{0:d}:{1:d}s}}  ".format(k, col_width)
                                for k in range(len(self.tags))])
        self.row_cls = row_factory(self.tags)
        self.data = [self.row_cls(row)
                     for (i, row) in enumerate(table) if i > 0]

    ##-------------------------------------------------------------------------
    def to_list(self):
        """
        Convert a Table to a list of lists with a header.
        """
        return [self.tags] + [row.values for row in self.data]

    ##-------------------------------------------------------------------------
    ##  support serialization

    def __getstate__(self):
        return self.to_list()

    def __setstate__(self, state):
        self.__init__(state)

    ##-------------------------------------------------------------------------
    def __len__(self):
        return len(self.data)

    ##-------------------------------------------------------------------------
    def __str__(self):
        return "\n".join(
            [self.row_fmt.format(*self.tags)] +
            [self.row_fmt.format(*["-"*len(tag) for tag in self.tags])] +
            [self.row_fmt.format(*row.str_values) for row in self.data])

    ##-------------------------------------------------------------------------
    def __getitem__(self, idx):
        """
        Get a row by index.
        """
        return self.data[idx]

    ##-------------------------------------------------------------------------
    def __eq__(self, other):
        return self.tags == other.tags and self.data == other.data

    ##-------------------------------------------------------------------------
    def append(self, values):
        self.data.append(self.row_cls(values))

    ##-------------------------------------------------------------------------
    def column(self, tag):
        """
        Description:
            Returns a list of values for the required column.
        Inputs:
            tag - the tag corresponding to the column to be returned
        Returns:
            A list with the required column values.
        """
        return [row.__getattribute__(tag) for row in self.data]

    ##-------------------------------------------------------------------------
    def sort(self, tags, reverse=False):
        """
        Description:
            Sort table in-place by specified tags.
        Inputs:
            tags    - a list of the tags corresponding to the columns to be
                      sorted
            reverse - if True, sorts in reverse order
        Returns:
            None.
        """
        self.data.sort(key=operator.attrgetter(*tags), reverse=reverse)

    ##-------------------------------------------------------------------------
    def find(self, tag, value):
        """
        Description:
            Find all the matching rows. The search is done on the specified
            column's tag.
        Inputs:
            tag   - the tag corresponding to the column to be searched
            value - the value
        Returns:
            A list of matching table_raw objects.
        """
        return [row for row in self.data if row.__getattribute__(tag) == value]



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################
"""
This module implements a Dependency Graph that works on top of the ObjDb
object database
"""

import onyx.database as onyx_db
import onyx.depgraph as onyx_dg

import weakref
import copy

__all__ = ["GraphError", "GraphNodeError"]


###############################################################################
class GraphError(Exception):
    pass


###############################################################################
class GraphNodeError(Exception):
    pass


###############################################################################
class DependencyGraph(dict):
    """
    Class representing the onyx Dependency Graph (a Directed Acyclic Graph).
    Current implementation uses a dictionary that maps (ObjectName, VT) tuples
    to graph nodes (instances of GraphNode class).
    """
    ##-------------------------------------------------------------------------
    def __init__(self):
        super().__init__()
        # --- hard references to the retrived objects are kept here so that
        #     any weak-references pointing to them won't disappear if and when
        #     the instance retrived from database is garbage collected
        self.objects = dict()
        # --- for each object, we keep the set of VTs that have been visited
        #     while building the graph
        self.vts_by_obj = dict()

    ##-------------------------------------------------------------------------
    def clear(self):
        super().clear()
        self.objects.clear()
        self.vts_by_obj.clear()


##-----------------------------------------------------------------------------
def CreateNode(node_def):
    """
    Description:
        Helper function used to create a new node.
    Inputs:
        node_def - the (ObjectName, VT) tuple
    Returns:
        A subclass of GraphNode.
    """
    obj_name, VT = node_def

    # --- the instance of the graph keeps trak of references to objects that
    #     have nodes in the current graph and maps VTs by object
    try:
        onyx_dg.graph.vts_by_obj[obj_name].add(VT)
    except KeyError:
        onyx_dg.graph.vts_by_obj[obj_name] = {VT}

    try:
        obj = onyx_dg.graph.objects[obj_name]
    except KeyError:
        obj = onyx_db.obj_clt.get(obj_name)
        onyx_dg.graph.objects[obj_name] = obj

    # --- all Instreams are settable by definition
    if VT in obj.Instreams:
        return GraphNodeSettable(VT, weakref.ref(obj))

    VT_descriptor = getattr(obj.__class__, VT)
    if hasattr(VT_descriptor, "fset"):
        return GraphNodeSettable(VT, weakref.ref(obj))
    elif callable(VT_descriptor):
        return GraphNodeCalc(VT, weakref.ref(obj))
    else:
        return GraphNode(VT, weakref.ref(obj))


###############################################################################
class GraphNode(object):
    """
    Base class representing a node in the Dependency Graph. Implements methods
    to get the node value (using memoization), to invalidate it (when
    recalculation is needed), and to clone the node (for diddling).
    This is used by ValueTypes implementing the property descriptor.
    """
    __slots__ = ("VT", "obj_ref", "parents",
                 "children", "valid", "value", "old_node")

    ##-------------------------------------------------------------------------
    def __init__(self, VT, obj_ref):
        self.VT = VT
        self.obj_ref = obj_ref
        self.parents = set()
        self.children = set()
        self.valid = False
        self.value = None
        self.old_node = None

    ##-------------------------------------------------------------------------
    def get_value(self):
        if self.valid:
            return self.value
        else:
            self.value = getattr(self.obj_ref(), self.VT)
            self.valid = True
            return self.value

    ##-------------------------------------------------------------------------
    def clone(self):
        cls = self.__class__
        new_node = cls.__new__(cls)
        for attr in self.__slots__:
            setattr(new_node, attr, copy.deepcopy(getattr(self, attr)))
        return new_node

    ##-------------------------------------------------------------------------
    def invalidate(self):
        if self.valid:
            # --- invalidate this node
            self.valid = False
            # --- recursively invalidate all parents
            for parent in self.parents:
                onyx_dg.graph[parent].invalidate()


###############################################################################
class GraphNodeSettable(GraphNode):
    """
    Derived class used for nodes whose value can be explicitly set (such as
    Instreams and descriptors that implement a setter).
    """
    __slots__ = ("VT", "obj_ref", "parents",
                 "children", "valid", "value", "old_node")


###############################################################################
class GraphNodeCalc(GraphNode):
    """
    Derived class used for calculated callable ValueTypes.
    """
    __slots__ = ("VT", "obj_ref", "parents", "children",
                 "valid", "value", "old_node", "args", "kwds")

    ##-------------------------------------------------------------------------
    def __init__(self, VT, obj_ref):
        super().__init__(VT, obj_ref)
        self.args = ()
        self.kwds = {}

    ##-------------------------------------------------------------------------
    def get_value(self, *args, **kwds):
        # --- if args and kwds don't match stored values, the graph node is
        #     assumed no longer valid
        if args != self.args or kwds != self.kwds:
            self.valid = False
            self.args = args
            self.kwds = kwds

        if self.valid:
            return self.value
        else:
            # --- call the object's method with provided arguments
            self.value = getattr(self.obj_ref(), self.VT)(*args, **kwds)
            self.valid = True
            return self.value



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.datatypes.curve import Curve, Knot
from onyx.datatypes.hlocv import HlocvCurve, HlocvKnot
from onyx.datatypes.date import Date
from onyx.database.objdb import typecast_to_Date

import psycopg2 as psql
import psycopg2.extras as psql_extras
import psycopg2.extensions as psql_ext
import sys

__all__ = ["TsNotFound", "TsDbError", "TsDbClient"]

# --- column names for HLOCVCurves
HLOCV_FIELDS = ("High", "Low", "Open", "Close", "Volume")

# --- define standard query strings
QUERY_TS_EXISTS = """
SELECT EXISTS (SELECT 1 FROM {0:s} WHERE Name=%s) AS "exists";"""


###############################################################################
class TsNotFound(Exception):
    pass


###############################################################################
class TsDbError(Exception):
    pass


##-----------------------------------------------------------------------------
def print_error(cursor):
    print("QUERY:\n{0:s}".format(cursor.query), file=sys.stderr)


###############################################################################
class TsDbClient(object):
    """
    Description:
        TsDb database, client class. It exposes a minimal api.
    """
    core_tables = ("Curves", "HLOCVCurves")

    ##-------------------------------------------------------------------------
    def __init__(self, database, user, host=None, check=True):
        self.dbname = database
        self.user = user
        self.host = host

        self.conn = psql.connect(host=host, database=database, user=user,
                                 cursor_factory=psql_extras.NamedTupleCursor)
        self.conn.set_isolation_level(psql_ext.ISOLATION_LEVEL_AUTOCOMMIT)

        # --- register typecast to Date for date, timpestamp, timestamptz
        curs = self.conn.cursor()
        curs.execute("SELECT NULL::date, NULL::timestamp, NULL::timestamptz;")
        oids = tuple([col.type_code for col in curs.description])
        psql_ext.register_type(psql_ext.new_type(oids, "Date",
                                                 typecast_to_Date))

        # --- validation: without the following tables in the backend the
        #                 client doesn't work properly.
        if check:
            curs = self.conn.cursor()
            for table in self.core_tables:
                try:
                    curs.execute("SELECT 1 FROM {0:s} LIMIT 1;".format(table))
                except psql.ProgrammingError:
                    raise RuntimeError("Table {0:s} is missing for "
                                       "{1:s}".format(table, self.dbname))

    ##-------------------------------------------------------------------------
    def initialize(self):
        curs = self.conn.cursor()
        curs.execute("""
            -- Curves table
            CREATE TABLE
            Curves (
                Date  timestamptz      NOT NULL,
                Name  varchar(64)      NOT NULL,
                Value double precision NOT NULL,
                CONSTRAINT CrvDateName UNIQUE (Date, Name) );

            CREATE INDEX crv_by_name_idx ON Curves (Name);

            -- HlocvCurves table
            CREATE TABLE
            HlocvCurves (
                Date   timestamptz      NOT NULL,
                Name   varchar(64)      NOT NULL,
                High   double precision NOT NULL,
                Low    double precision NOT NULL,
                Open   double precision NOT NULL,
                Close  double precision NOT NULL,
                Volume double precision NOT NULL,
                CONSTRAINT HlocvDateName UNIQUE (Date, Name) );

            CREATE INDEX hlocv_by_name_idx ON HlocvCurves (Name);""")

    ##-------------------------------------------------------------------------
    def restart(self):
        self.close()
        self.conn = psql.connect(host=self.host,
                                 database=self.dbname, user=self.user)
        self.conn.set_isolation_level(psql_ext.ISOLATION_LEVEL_AUTOCOMMIT)

    ##-------------------------------------------------------------------------
    def close(self):
        if hasattr(self, "conn") and not self.conn.closed:
            self.conn.close()

    ##-------------------------------------------------------------------------
    def get_row_by(self, table, name, date=None, strict=False):
        if date is None:
            query = """SELECT * FROM {0:s} WHERE Name=%s AND Date IN
                       (SELECT MAX(Date) FROM {0:s} WHERE Name=%s);"""
            parms = (name, name)
        elif strict:
            query = "SELECT * FROM {0:s} WHERE Name=%s AND Date=%s;"
            parms = (name, date)
        else:
            query = """SELECT * FROM {0:s} WHERE Name=%s AND Date IN
                       (SELECT MAX(Date)
                        FROM {0:s} WHERE Name=%s AND Date<=%s);"""
            parms = (name, name, date)

        curs = self.conn.cursor()
        try:
            curs.execute(query.format(table), parms)
        except psql.OperationalError:
            print_error(curs)
            raise

        row = curs.fetchone()
        curs.close()

        if row is None:
            if strict and date is not None:
                raise TsNotFound("Row not found for curve {0:s} in {1:s}"
                                 "and date {2:s}".format(name, table, date))
            else:
                raise TsNotFound("No rows were found "
                                 "for {0:s} in {1:s}".format(name, table))
        else:
            return row

    ##-------------------------------------------------------------------------
    def get_curve(self, name, start=None, end=None):
        if start is None:
            if end is None:
                # --- neither start nor end are specified
                parms = name,
                query = "SELECT Date, Value FROM Curves WHERE Name=%s;"
                errmsg = "No data found in Curves for {0:s}".format(name)
            else:
                # --- end is specified, start is not
                parms = name, end
                query = """SELECT Date, Value FROM Curves
                           WHERE Name = %s AND Date <= %s;"""
                errmsg = "No data found in " \
                         "Curves for {0:s}, end = {1:s}".format(name, end)
        elif end is None:
            # --- start is specified, end is not
            parms = name, start
            query = """SELECT Date, Value FROM Curves
                       WHERE Name = %s AND Date >= %s;"""
            errmsg = "No data found in Curves " \
                     "for {0:s}, start = {1:s}".format(name, start)
        else:
            # --- both start and end are specified
            parms = name, start, end
            query = """SELECT Date, Value
                       FROM Curves
                       WHERE Name = %s AND Date BETWEEN %s AND %s;"""
            errmsg = "No data found in Curves for {0:s}, " \
                     "start = {1:s} end = {2:s}".format(name, start, end)

        # --- create a cursor, check if ts exists and execute query
        curs = self.conn.cursor()
        try:
            curs.execute(QUERY_TS_EXISTS.format("Curves"), (name, ))
        except psql.OperationalError:
            print_error(curs)
            raise

        if not curs.fetchone().exists:
            raise TsNotFound("Time Series {0:s} not "
                             "found in Curves table".format(name))

        try:
            curs.execute(query, parms)
        except psql.OperationalError:
            print_error(curs)
            raise
        else:
            # --- populate a curve with the query's results
            crv = Curve.__new__(Curve)
            crv.knot_cls = Knot
            crv.data = [Knot(Date.parse(r[0]), r[1]) for r in curs.fetchall()]
            crv.data.sort()
        finally:
            curs.close()

        # --- raise exception if curve is empty
        if not len(crv):
            raise TsNotFound(errmsg)

        return crv

    ##-------------------------------------------------------------------------
    def get_hlocv(self, name, start=None, end=None, field=None):
        if start is None:
            if end is None:
                # --- neither start nor end are specified
                parms = name,
                query = """SELECT
                           Date, High, Low, Open, Close, Volume
                           FROM HLOCVCurves
                           WHERE Name = %s;"""
                errmsg = "No data found in HLOCVCurves for {0:s}".format(name)
            else:
                # --- end is specified, start is not
                parms = name, end
                query = """SELECT
                           Date, High, Low, Open, Close, Volume
                           FROM HLOCVCurves
                           WHERE Name = %s AND Date <= %s;"""
                errmsg = "No data found in " \
                         "HLOCVCurves for {0:s}, end = {1:s}".format(name, end)
        elif end is None:
            # --- start is specified, end is not
            parms = name, start
            query = """SELECT
                       Date, High, Low, Open, Close, Volume
                       FROM HLOCVCurves
                       WHERE Name = %s AND Date >= %s;"""
            errmsg = "No data found in " \
                     "HLOCVCurves for {0:s}, start = {1:s}".format(name, start)
        else:
            # --- both start and end are specified
            parms = name, start, end
            query = """SELECT
                       Date, High, Low, Open, Close, Volume
                       FROM HLOCVCurves
                       WHERE Name = %s AND Date BETWEEN %s AND %s;"""
            errmsg = "No data found in HLOCVCurves for {0:s}, " \
                     "start = {1:s} end = {2:s}".format(name, start, end)

        # --- create a cursor, check if ts exists and execute query
        curs = self.conn.cursor()
        try:
            curs.execute(QUERY_TS_EXISTS.format("HLOCVCurves"), (name, ))
        except psql.OperationalError:
            print_error(curs)
            raise
        if not curs.fetchone().exists:
            raise TsNotFound("Time Series {0:s} not found "
                             "in HLOCVCurves table".format(name))

        try:
            curs.execute(query, parms)
        except psql.OperationalError:
            print_error(curs)
            raise
        else:
            rows = curs.fetchall()
            if field is None:
                crv = HlocvCurve.__new__(HlocvCurve)
                crv.knot_cls = HlocvKnot
                crv.data = [HlocvKnot(Date.parse(r[0]), r[1:]) for r in rows]
            else:
                idx = 1 + HLOCV_FIELDS.index(field)
                crv = Curve.__new__(Curve)
                crv.knot_cls = Knot
                crv.data = [Knot(Date.parse(r[0]), r[idx]) for r in rows]
            crv.data.sort()
        finally:
            curs.close()

        # --- raise exception if curve is empty
        if not len(crv):
            raise TsNotFound(errmsg)

        return crv

    ##-------------------------------------------------------------------------
    def upsert_curve(self, name, crv, crv_type="CRV"):
        if crv_type == "HLOCV":
            upd_unpack = lambda k: (k.high, k.low, k.open,
                                    k.close, k.volume, k.date, name)
            ins_unpack = lambda k: (k.date, name, k.high,
                                    k.low, k.open, k.close, k.volume)
        else:
            upd_unpack = lambda k: (k.value, k.date, name)
            ins_unpack = lambda k: (k.date, name, k.value)

        ins = []
        upd = []

        try:
            # --- get existing curve
            if crv_type == "HLOCV":
                curr_crv = self.get_hlocv(name, crv.front.date, crv.back.date)
            else:
                curr_crv = self.get_curve(name, crv.front.date, crv.back.date)
        except TsNotFound:
            ins = [ins_unpack(knot) for knot in crv]
        else:
            # --- loop over all knots and check which one needs updating
            #     and which one needs inserting
            for knot in crv:
                try:
                    curr_knot = curr_crv.get_knot(knot.date)
                    # --- knot exists already, update if value has changed
                    if curr_knot.value != knot.value:
                        upd.append(upd_unpack(knot))
                except IndexError:
                    # --- knot not found, insert
                    ins.append(ins_unpack(knot))

        if crv_type == "HLOCV":
            ins_query = "INSERT INTO HlocvCurves VALUES (%s,%s,%s,%s,%s,%s,%s)"
            upd_query = """UPDATE HlocvCurves
                           SET
                           High = %s,
                           Low = %s,
                           Open = %s,
                           Close = %s,
                           Volume = %s
                           WHERE Date = %s AND Name = %s"""
        else:
            ins_query = "INSERT INTO Curves VALUES (%s,%s,%s)"
            upd_query = """UPDATE Curves
                           SET
                           Value = %s
                           WHERE Date = %s AND Name = %s"""

        curs = self.conn.cursor()

        if len(ins):
            curs.executemany(ins_query, ins)
        if len(upd):
            curs.executemany(upd_query, upd)



###############################################################################
#
#   Onyx Portfolio & Risk Management Framework
#
#   Copyright 2014 Carlo Sbraccia
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
###############################################################################

from onyx.datatypes.table import row_factory, Table
from onyx.database.ufo_base import UfoBase
from onyx.database.ufo_fields import TableField, DictField
from onyx.database.objdb_api import ObjDbQuery

__all__ = ["TableSchema", "ObjDbTable"]

TAGS = ("column_name", "data_type", "column_constraint")
ROW_CLS = row_factory(TAGS)


###############################################################################
class TableSchema(Table):
    ##-------------------------------------------------------------------------
    def __init__(self, schema):
        self.tags = TAGS
        self.row_fmt = "{0:20s}  {1:20s}  {2:20s}"
        self.row_cls = ROW_CLS
        self.data = [ROW_CLS(row) for (i, row) in enumerate(schema)]


###############################################################################
class ObjDbTable(UfoBase):
    """
    This UFO class is used to crerate and manage backend database tables.
    """
    Schema = TableField()
    Unique = DictField({})
    Indexes = DictField({})

    ##-------------------------------------------------------------------------
    def __post_init__(self):
        # --- validate indexes
        for index_name, index_columns in self.Indexes.items():
            if not set(index_columns).issubset(self.columns):
                raise ValueError("Index {0:s} is not valid (some "
                                 "of the columns are not defined "
                                 "for this table's schema)".format(index_name))

    ##-------------------------------------------------------------------------
    @property
    def columns(self):
        return [row.column_name for row in self.Schema]

    ##-------------------------------------------------------------------------
    @property
    def __create_table(self):
        col_fmt = "    {0:s},\n"
        sql = ["CREATE TABLE {0:s} (\n".format(self.Name)]
        for entry in self.Schema:
            sql.append(col_fmt.format(" ".join(entry.str_values).rstrip()))

        for constr_name, constr_columns in self.Unique.items():
            columns = ",".join(constr_columns)
            sql.append("    CONSTRAINT {0:s} UNIQUE "
                       "({1:s}),\n".format(constr_name, columns))

        # --- strip the last comma
        sql[-1] = sql[-1].replace(",\n", "\n")

        sql.append(");")
        return "".join(sql)

    ##-------------------------------------------------------------------------
    @property
    def __create_indexes(self):
        sql = []
        for index_name, index_columns in self.Indexes.items():
            columns = ",".join(index_columns)
            sql.append("CREATE INDEX {0:s} ON {1:s} "
                       "({2:s});\n".format(index_name, self.Name, columns))
        return "".join(sql)

    ##-------------------------------------------------------------------------
    def create(self):
        ObjDbQuery(self.__create_table)
        ObjDbQuery(self.__create_indexes)

    ##-------------------------------------------------------------------------
    def delete(self):
        # --- drop table from backend
        ObjDbQuery("DROP TABLE IF EXISTS {0:s} CASCADE;".format(self.Name))


if __name__ == "__main__":
    schema = TableSchema([
        ["TradeDate", "date", "NOT NULL"],
        ["TimeCreated", "timestamptz", "NOT NULL"],
        ["Book", "varchar(64)", "NOT NULL"],
        ["Trader", "varchar(64)", "NOT NULL"],
        ["Qty", "double precision", "NOT NULL"],
        ["Unit", "varchar(64)", "NOT NULL"],
        ["UnitType", "varchar(64)", "NOT NULL"],
        ["Status", "char(1)", "NOT NULL"],
        ["Trade", "varchar(64)", "NOT NULL"],
        ["DealId", "varchar(64)", "NOT NULL"],
        ["NTD", "date", ""]
    ])

    unique = {
        "tarde_and_status": ["Trade", "Status"],
    }

    indexes = {
        "pos_by_trade_date": ["TradeDate"],
        "pos_by_time_created": ["TimeCreated"],
        "pos_by_book": ["Book"],
        "pos_by_trader": ["Trader"],
        "pos_by_trade": ["DealId"],
        "pos_by_ntd": ["NTD"],
    }

    table = ObjDbTable(Name="PosEffects", Schema=schema,
                       Unique=unique, Indexes=indexes)

    print(table._ObjDbTable__create_table)
